from __future__ import annotations

from typing import Optional
from collections import UserList
from dataclasses import dataclass, field
from wxcadm import log
from .common import *


class Webhooks(UserList):
    def __init__(self):
        log.info("Initializing Webhooks instance")
        super().__init__()
        self.data = []
        items = webex_api_call("get", "v1/webhooks")
        log.debug(f"Webex returned {len(items)} webhooks")
        for item in items:
            webhook = Webhook(**item)
            self.data.append(webhook)

    @property
    def active(self):
        """ Return only the active Webhook instances """
        active = []
        for webhook in self.data:
            if webhook.status.lower() == "active":
                active.append(webhook)
        return active

    @property
    def inactive(self):
        """ Return only the inactive Webhook instances """
        inactive = []
        for webhook in self.data:
            if webhook.status.lower() == "inactive":
                inactive.append(webhook)
        return inactive

    def add(self, name: str,
            url: str,
            resource: str,
            event: str,
            filter: str = None,
            secret: str = None,
            owner: str = None):
        """ Add a new Webhook

        Possible values for arguments can be found at `<https://developer.webex.com/docs/webhooks>`_

        Args:
            name (str): The name of the Webhook
            url (str): The URL to which the Webbook mesages should be sent
            resource (str): The resource type of the Webhook
            event (str): The event type to trigger the Webhook
            filter (str, optional): A valid filter for the Webhook, if needed
            secret (str, optional): The secret value used to generate the Webhook signature
            owner (str, optional): The Webhook owner is creating an org-wide Webhook

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Creating Webhook with name: {name}")
        payload = {"name": name,
                   "targetUrl": url,
                   "resource": resource,
                   "event": event,
                   "filter": filter,
                   "secret": secret,
                   "ownedBy": owner}
        new_webhook = webex_api_call("post", "v1/webhooks", payload=payload)
        if new_webhook:
            log.debug(f"New Webhook ID: {new_webhook['id']}")
            self.data.append(Webhook(**new_webhook))
            return True
        else:
            log.warning("The Webhook creation failed")
            return False


@dataclass
class Webhook:
    """ The Webhook class contains information about each Webhook """
    orgId: str = field(repr=False)
    appId: str
    """ The ID of the application """
    id: str
    """ The unique ID of the Webhook """
    name: str
    """ The user-defined name of the Webhook """
    targetUrl: str
    """ The URL to which messages will be sent """
    resource: str
    """ The resource type for the Webhook """
    event: str
    """ The event type (e.g. created, updated, etc...) for the Webhook """
    status: str
    """ The status of the Webhook, either active or inactive """
    created: str
    """ The date and time the Webhook was created """
    createdBy: str
    """ The ID of the user who created the Webhook """
    ownedBy: str
    """ The owner of the Webhook. Specified when creating an org-level Webhook """
    filter: Optional[str] = None
    """ Any filter that is applied to the Webhook """
    secret: Optional[str] = None
    """ The secret used to generate the payload signature """

    def delete(self) -> bool:
        """ Delete the Webhook

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Deleting Webhook: {self.name}")
        success = webex_api_call("delete", f"v1/webhooks/{self.id}")
        if success:
            return True
        else:
            log.warning("The Webhook delete failed")
            return False

    def change_url(self, url: str) -> bool:
        """ Change the URL to which the Webhook will be sent

        Args:
            url (str): The new URL for the Webhook

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Changing Webhook URL for {self.name} to {url}")
        payload = {"name": self.name,
                   "targetUrl": url,
                   "secret": self.secret,
                   "ownedBy": self.ownedBy,
                   "status": self.status}
        success = webex_api_call("put", f"v1/webhooks/{self.id}", payload=payload)
        if success:
            self.targetUrl = url
            return True
        else:
            log.warning("The Webhook change failed")
            return False

    def change_name(self, name: str) -> bool:
        """ Change the name of the Webhook

        Args:
            name (str): The new name for the Webhook

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Changing Webhook name for {self.name} to {name}")
        payload = {"name": name,
                   "targetUrl": self.targetUrl,
                   "secret": self.secret,
                   "ownedBy": self.ownedBy,
                   "status": self.status}
        success = webex_api_call("put", f"v1/webhooks/{self.id}", payload=payload)
        if success:
            self.name = name
            return True
        else:
            log.warning("The Webhook change failed")
            return False

    def change_secret(self, secret: str) -> bool:
        """ Change the secret value used to generate the Webhook signature

        Args:
            secret (str): The new secret value

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Changing Webhook secret for {self.name} to {secret}")
        payload = {"name": self.name,
                   "targetUrl": self.targetUrl,
                   "secret": secret,
                   "ownedBy": self.ownedBy,
                   "status": self.status}
        success = webex_api_call("put", f"v1/webhooks/{self.id}", payload=payload)
        if success:
            self.secret = secret
            return True
        else:
            log.warning("The Webhook change failed")
            return False

    def change_owner(self, owner: str) -> bool:
        """ Change the owner of the Webhook

        Args:
            owner (str): The new owner value

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Changing Webhook owner for {self.name} to {owner}")
        payload = {"name": self.name,
                   "targetUrl": self.targetUrl,
                   "secret": self.secret,
                   "ownedBy": owner,
                   "status": self.status}
        success = webex_api_call("put", f"v1/webhooks/{self.id}", payload=payload)
        if success:
            self.ownedBy = owner
            return True
        else:
            log.warning("The Webhook change failed")
            return False

    def deactivate(self) -> bool:
        """ Change the status of the Webhook to inactive

        Returns:
            bool: True on success, False otherwise. True will be returned if the Webhook is already inactive.

        """
        log.info(f"Deactivating Webhook: {self.name}")
        payload = {"name": self.name,
                   "targetUrl": self.targetUrl,
                   "secret": self.secret,
                   "ownedBy": self.ownedBy,
                   "status": "inactive"}
        success = webex_api_call("put", f"v1/webhooks/{self.id}", payload=payload)
        if success:
            self.status = "inactive"
            return True
        else:
            log.warning("The Webhook change failed")
            return False

    def activate(self) -> bool:
        """ Change the status of the Webhook to active

        Returns:
            bool: True on success, False otherwise. True will be returned if the Webhook is already active.

        """
        log.info(f"Activating Webhook: {self.name}")
        payload = {"name": self.name,
                   "targetUrl": self.targetUrl,
                   "secret": self.secret,
                   "ownedBy": self.ownedBy,
                   "status": "active"}
        success = webex_api_call("put", f"v1/webhooks/{self.id}", payload=payload)
        if success:
            self.status = "active"
            return True
        else:
            log.warning("The Webhook change failed")
            return False



from __future__ import annotations

from collections import UserList
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from typing import Optional, Union, List

import wxcadm
import wxcadm.location
import wxcadm.person
from wxcadm import log
from .common import *

@dataclass_json
@dataclass
class MonitoringList:
    parent: Union[wxcadm.Person, wxcadm.Workspace]
    org: wxcadm.Org
    call_park_notification: bool = field(metadata=config(field_name="callParkNotificationEnabled"))
    """ Whether to notify the user about calls parked on monitored lines """
    monitored_elements: list = field(metadata=config(field_name="monitoredElements"))
    """ List of monitored elements, which may be Person, Workspace, CallParkExtension or VirtualLine """
    _url: str = field(init=False, repr=False)

    def __post_init__(self):
        # Set the _url based on whether this is for a Person or a Workspace
        if isinstance(self.parent, wxcadm.Workspace):
            self._url = f"v1/workspaces/{self.parent.id}/features/monitoring"
        else:
            self._url = f"v1/people/{self.parent.id}/features/monitoring"
        log.debug("Resolving monitored elements")
        new_monitoring_list = []
        for monitored_element in self.monitored_elements:
            element_type, element_info = list(monitored_element.items())[0]
            if element_type == 'callparkextension':
                park_extensions = self.org.locations.get(id=element_info['locationId']).park_extensions
                for park_extension in park_extensions:
                    if park_extension.id == element_info['id']:
                        this_element = park_extension
                        break
            elif element_type == 'member':
                if element_info['type'] == 'PEOPLE':
                    this_element = self.org.people.get(id=element_info['id'])
                elif element_info['type'] == 'PLACE':
                    this_element = self.org.workspaces.get(id=element_info['id'])
                elif element_info['type'] == 'VIRTUAL_LINE':
                    this_element = self.org.virtual_lines.get(id=element_info['id'])
                else:
                    log.warning("Unknown element info type: {}".format(element_info['type']))
                    log.debug(f"Details: {element_info}")
                    continue
            else:
                log.warning("Unknown element type: {}".format(element_type))
                continue
            new_monitoring_list.append(this_element)
        self.monitored_elements = new_monitoring_list

    def add(self, monitor: Union[wxcadm.Person, wxcadm.VirtualLine, wxcadm.Workspace, wxcadm.CallParkExtension]):
        """ Add a new monitoring

        Args:
            monitor (Person, VirtualLine, Workspace, CallParkExtension): The new element to monitor

        Returns:
            bool: True on success, False otherwise

        """
        payload = {
            "enableCallParkNotification": self.call_park_notification,
            "monitoredElements": [],
        }
        for element in self.monitored_elements:
            payload["monitoredElements"].append(element.id)
        payload["monitoredElements"].append(monitor.id)
        webex_api_call("put", url=self._url, payload=payload,
                       params={"orgId": self.org.id})
        self.monitored_elements.append(monitor)
        return True

    def remove(self, monitor: Union[wxcadm.Person, wxcadm.VirtualLine, wxcadm.Workspace, wxcadm.CallParkExtension]):
        """ Remove a monitoring

        Args:
            monitor: (Person, VirtualLine, Workspace, CallParkExtension): The new element to monitor

        Returns:
            bool: True on success, False otherwise

        """
        payload = {
            "enableCallParkNotification": self.call_park_notification,
            "monitoredElements": [],
        }
        new_monitoring_list = []
        for element in self.monitored_elements:
            if element.id != monitor.id:
                payload["monitoredElements"].append(element.id)
                new_monitoring_list.append(element)
        webex_api_call("put", url=self._url, payload=payload,
                       params={"orgId": self.org.id})
        self.monitored_elements = new_monitoring_list
        return True

    def replace(self, monitoring_list: MonitoringList):
        """ Replace the monitoring config and monitored elements with a new :class:`~.monitoring.MonitoringList`

        Args:
            monitoring_list (MonitoringList): The new monitoring list

        Returns:
            bool: True on success, False otherwise

        """
        payload = {
            "enableCallParkNotification": monitoring_list.call_park_notification,
            "monitoredElements": [],
        }
        for element in monitoring_list.monitored_elements:
            # Remove the current User/Workspace if it is in the list
            if element.id != self.parent.id:
                payload["monitoredElements"].append(element.id)
        webex_api_call("put", url=self._url, payload=payload,
                       params={"orgId": self.org.id})
        self.monitored_elements = monitoring_list.monitored_elements
        self.call_park_notification = monitoring_list.call_park_notification
        return True

    def copy_to(self, target: Union[wxcadm.Person, wxcadm.Workspace]):
        """ Copy the monitoring config to another Person or Workspace

        Args:
            target (Person, Workspace): Person or Workspace to copy to

        Returns:
            bool: True on success, False otherwise

        """
        target.monitoring.replace(self)
        return True

    def copy_from(self, source: Union[wxcadm.Person, wxcadm.Workspace]):
        """ Copy the monitoring config from another Person or Workspace

        Args:
            source (Person, Workspace): Person or Workspace to copy from

        Returns:
            bool: True on success, False otherwise

        """
        self.replace(source.monitoring)
        return True

    def clear(self):
        """ Clear the list of monitored elements. This removes all monitoring.

        Returns:
            bool: True on success, False otherwise

        """
        log.debug(f"Clearing monitoring list for {self.parent.id}")
        payload = {
            "enableCallParkNotification": False,
            "monitoredElements": [],
        }
        webex_api_call("put", url=self._url, payload=payload, params={"orgId": self.org.id})
        self.monitored_elements = []
        self.call_park_notification = False
        return True




import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from wxcadm.location import *
    from wxcadm.location_features import *


class TestLocationECBN(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location: Location = choice(self.webex.org.locations.webex_calling())

    def test_location_ecbn_get(self) -> None:
        current_ecbn = self.random_location.ecbn
        self.assertIsInstance(current_ecbn, dict)
        self.assertIn(current_ecbn['selected'], ['LOCATION_NUMBER', 'LOCATION_MEMBER_NUMBER'], "Unknown selected value")


class TestVoicemailGroups(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location: Location = choice(self.webex.org.locations.webex_calling())

    def test_org_voicemail_group_list(self) -> None:
        self.assertIsInstance(self.webex.org.voicemail_groups, wxcadm.VoicemailGroupList)

    def test_create_update_delete_vm_group(self) -> None:
        before_count = len(self.webex.org.voicemail_groups)
        with self.subTest('Add Voicemail Group'):
            new_group = self.webex.org.voicemail_groups.create(
                location=self.random_location,
                name='wxcadm Test',
                extension='928807',
                passcode='288547'
            )
            self.assertIsInstance(new_group, wxcadm.VoicemailGroup)
            self.assertEqual(len(self.webex.org.voicemail_groups.refresh()), before_count + 1)

        with self.subTest('Update VM Group Email'):
            new_group.enable_email_copy('wxcadmautotest@wxcadm.com')
            self.assertTrue(new_group.email_copy_of_message['enabled'])
            self.assertEqual(new_group.email_copy_of_message['emailId'], 'wxcadmautotest@wxcadm.com')
            new_group.update(first_name='Test')
            self.assertEqual(new_group.first_name, 'Test')

        with self.subTest('Delete VM Group'):
            new_group.delete()
            self.assertEqual(len(self.webex.org.voicemail_groups.refresh()), before_count)


class TestOutgoingDigitPatterns(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location: Location = choice(self.webex.org.locations.webex_calling())

    def test_outgoing_permission_digit_pattern_list(self) -> None:
        self.assertIsInstance(self.random_location.outgoing_permission_digit_patterns,
                              wxcadm.OutgoingPermissionDigitPatternList)

    def test_add_update_remove_pattern(self) -> None:
        # To confirm we are adding and deleting, get the len of the current list for the location
        before_len = len(self.random_location.outgoing_permission_digit_patterns.patterns)
        with self.subTest('Add digit pattern'):
            new_pattern: OutgoingPermissionDigitPattern = \
                self.random_location.outgoing_permission_digit_patterns.create(
                    name='wxcadm Test',
                    pattern='012XXX',
                    action='ALLOW'
                )
            self.assertIsInstance(new_pattern, wxcadm.OutgoingPermissionDigitPattern)
            self.assertEqual(
                len(self.random_location.outgoing_permission_digit_patterns.patterns),
                before_len + 1
            )
        with self.subTest('Update pattern attributes'):
            # Set each value individually
            new_pattern.update(name='wxcadm Change Test')
            self.assertEqual(new_pattern.name, 'wxcadm Change Test')
            new_pattern.update(pattern='013XXX')
            self.assertEqual(new_pattern.pattern, '013XXX')
            new_pattern.update(action='BLOCK')
            self.assertEqual(new_pattern.action, 'BLOCK')
            new_pattern.update(transfer_enabled=True)
            self.assertEqual(new_pattern.transfer_enabled, True)
            # Set all values at one time
            new_pattern.update(name='wxcadm Final Test', action='ALLOW', pattern='014XXX', transfer_enabled=False)
        with self.subTest('Delete pattern'):
            success = new_pattern.delete()
            self.assertEqual(success, True)
            self.assertEqual(
                len(self.random_location.outgoing_permission_digit_patterns.refresh().patterns),
                before_len
            )


class TestLocationFloors(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location: Location = choice(self.webex.org.locations.webex_calling())

    def test_floor_list(self) -> None:
        self.assertIsInstance(self.random_location.floors, wxcadm.LocationFloorList)
        if len(self.random_location.floors) == 0:
            self.skipTest("No floors at location")
        self.assertIsInstance(self.random_location.floors[0], wxcadm.LocationFloor)
        self.assertIsInstance(self.random_location.floors[0].floor_number, int)


if __name__ == '__main__':
    unittest.main()


from typing import Optional, Union

import wxcadm
from wxcadm import log
from .redsky import RedSky
import meraki


def tags_decoder(tags: list, match_string: Optional[str] = "911-") -> Optional[str]:
    """ Takes a list of tags and finds the one matching the ``match_string``

    When the match is found, it also converts underscores to spaces and returns the value without the ``match_string``.

    Args:
        tags (list): A list of tags, normally from a Meraki device
        match_string (str): A string to match to find the first tag. Defaults to ``'911-'``

    Returns:
        str: The normalized value of the matching tag.

    """
    loc_tag = None
    for tag in tags:
        if match_string in tag:
            loc_tag = tag
            break
    if loc_tag is None:
        return None
    loc_tag = loc_tag.replace(match_string, "")
    loc_tag = loc_tag.replace("_", " ")
    return loc_tag


def address_cleaner(address: str) -> str:
    clean_address = address.replace("\n", ", ")
    return clean_address


class Meraki:
    """ The top-level class for communicating with Meraki """
    def __init__(self, api_key: Optional[str] = None):
        """ Initialize a Meraki instance

        Args:
            api_key (str): The API key provided by Meraki to manage the Dashboard via API

        """

        log.info("Connecting to the Meraki Dashboard")
        self.dashboard = meraki.DashboardAPI(api_key=api_key, suppress_logging=True)
        """ The API connection to the Meraki Dashboard """
        self._orgs = []

    def get_orgs(self):
        """ Get the list of :py:class:`MerakiOrg` instances that are managed by the API token

        Returns:
            list[MerakiOrg]

        """
        log.info("Getting list of organizations from Meraki")
        org_list = self.dashboard.organizations.getOrganizations()
        log.debug(f"\t{org_list}")
        for org in org_list:
            this_org = MerakiOrg(self.dashboard, org)
            self._orgs.append(this_org)
        return self._orgs

    def get_org_by_name(self, name: str):
        """ Get the :py:class:`MerakiOrg` instance for a specific Organization name

        Args:
            name (str): The name of the Meraki Organization to get

        Returns:
            MerakiOrg: The :class:`MerakiOrg` instance

        """
        log.info(f"Getting organization with name: {name}")
        if not self._orgs:
            self.get_orgs()
        org: MerakiOrg
        for org in self._orgs:
            if org.name.upper() == name.upper():
                return org
        return None


class MerakiOrg:
    def __init__(self, dashboard: meraki.DashboardAPI, org_config: dict):
        self.dashboard = dashboard
        """ The API connection to the Meraki dashboard """
        self.id = org_config.get('id', None)
        """ The ID of the organization """
        self.name = org_config.get('name', None)
        """ The organization name """
        self._networks = []

    def get_networks(self):
        """ Get all the :class:`MerakiNetwork` instances for the Organization

        Returns:
            list[MerakiNetwork]: A list of the MerakiNetwork instances

        """
        log.info(f"Getting networks for Organization ID {self.id}")
        network_list = self.dashboard.organizations.getOrganizationNetworks(organizationId=self.id)
        log.debug(f"\t{network_list}")
        for network in network_list:
            this_network = MerakiNetwork(self.dashboard, network)
            self._networks.append(this_network)
        return self._networks

    def get_network_by_name(self, name: str):
        """ Get the :py:class:`MerakiNetwork` instance for a specified Network name

        Args:
            name (str): The Network name to get

        Returns:
            MerakiNetwork: The matching instance

        """
        log.info(f"Getting Network with name: {name}")
        if not self._networks:
            self.get_networks()
        net: MerakiNetwork
        for net in self._networks:
            if net.name.upper() == name.upper():
                return net
        return None


class MerakiNetwork:
    def __init__(self, dashboard: meraki.DashboardAPI, network_config: dict):
        self.dashboard = dashboard
        """ The API connection to the Merak Dashboard """
        self.id = network_config.get('id', None)
        """ The Network ID """
        self.name = network_config.get('name', None)
        """ The name of the Network """
        self._devices = []
        self._redsky: Optional[RedSky] = None

    def get_devices(self):
        """ Get a list of all :py:class:`MerakiDevice` instances for the Network

        Returns:
            list[MerakiDevice]: List of the MerakiDevice instances

        """
        device_list = self.dashboard.networks.getNetworkDevices(self.id)
        for device in device_list:
            if "MS" in device['model']:
                this_device = MerakiSwitch(self.dashboard, device)
            elif "MR" in device['model']:
                this_device = MerakiWireless(self.dashboard, device)
            else:
                continue
            if self._redsky is not None:
                this_device._redsky = self._redsky
            self._devices.append(this_device)
        return self._devices

    def attach_redsky(self, username: str, password: str) -> bool:
        """ Attach a :py:class:`wxcadm.RedSky` instance to the Meraki Network

        Args:
            username (str): The RedSky admin user
            password (str): The RedSky admin password

        Returns:
            bool: True on success, False otherwise

        """
        try:
            redsky_connection = RedSky(username, password)
        except wxcadm.APIError:
            return False
        self._redsky = redsky_connection
        for device in self._devices:
            device._redsky = redsky_connection
        return True

    def redsky_audit(self, simulate: bool = False):
        """ Audit the Meraki network against RedSky Horizon Mobility.

        Network devices with a tag that begins with "911-" will be used to build the RedSky Network Discovery wiremap,
        including Locations. The audit assumes that the Buildings already exist, and the address is used to match.

        Args:
            simulate (bool, optional): When set to True, no changes will be performed on the RedSky platform.
                The AuditResults will show the changes that would have been made. Defaults to False

        Returns:
            MerakiAuditResults: The results of the network audit

        """
        if self._redsky is None:
            raise RuntimeError("No RedSky instance attached")

        audit_results = MerakiAuditResults()
        device: Union[MerakiDevice, MerakiSwitch, MerakiWireless]
        for device in self.get_devices():
            log.info(f"Auditing device: {device.name}")
            device_building: wxcadm.redsky.RedSkyBuilding = device.get_redsky_building()
            if device_building is None:
                audit_results.missing_buildings.append(device.address)
                # new_building = self._redsky.add_building(address_string=device.address)
                continue

            if isinstance(device, MerakiSwitch):
                audit_results.switches.append(device)
                # Chassis
                device_location_tag = tags_decoder(device.tags)
                if device_location_tag is None:
                    log.info("No location information in Meraki. Skipping")
                    audit_results.ignored_devices.append(device)
                    continue
                # Get the RedSky Location within the building that corresponds to the location tag
                chassis_redsky_location = device_building.get_location_by_name(name=device_location_tag)
                if chassis_redsky_location is None:
                    log.info(f"No Location found for {device_location_tag} in Building {device_building.name}. Adding.")
                    if simulate is False:
                        device_building.add_location(location_name=device_location_tag,
                                                     location_info=device_location_tag)
                        chassis_redsky_location = device_building.get_location_by_name(name=device_location_tag)
                    audit_results.added_locations.append({'building': device_building, 'location': device_location_tag})

                chassis_lldp_discovery = self._redsky.get_lldp_discovery_by_chassis(chassis_id=device.mac)
                if chassis_lldp_discovery is None:
                    log.info(f"Chassis {device.mac} not in RedSky. Adding.")
                    if simulate is False:
                        self._redsky.add_lldp_discovery(chassis=device.mac,
                                                        location=chassis_redsky_location,
                                                        description=device.name)
                    audit_results.added_chassis.append(device)
                else:
                    # Check the existing LLDP discovery location against what Meraki says it should be
                    if chassis_lldp_discovery['erl']['name'] != device_location_tag:
                        log.info(f"Chassis {device.mac} has wrong location. Updating to {device_location_tag}")
                        if simulate is False:
                            self._redsky.update_lldp_location(entry_id=chassis_lldp_discovery['id'],
                                                              chassis=device.mac,
                                                              new_location=chassis_redsky_location,
                                                              description=device.name)
                        audit_results.updated_locations.append(device)


                # Ports
                port: MerakiSwitchPort
                for port in device.get_ports():
                    port_location_tag = tags_decoder(port.tags)
                    if port_location_tag is None:
                        # The port uses the chassis location. No work required
                        continue
                    port_redsky_location = device_building.get_location_by_name(name=port_location_tag)
                    if port_redsky_location is None:
                        log.info(f"No Location found for {port_location_tag} in Building {device_building.name}. "
                                 f"Adding.")
                        if simulate is False:
                            device_building.add_location(location_name=port_location_tag,
                                                         location_info=port_location_tag)
                            port_redsky_location = device_building.get_location_by_name(name=port_location_tag)
                        audit_results.added_locations.append({'building': device_building,
                                                              'location': port_location_tag})

                    chassis_lldp_discovery = self._redsky.get_lldp_discovery_by_chassis(chassis_id=device.mac)
                    # If we are in simulate mode, there may not be a chassis LLDP, so we have to fake it
                    if simulate is True:
                        chassis_lldp_discovery = {'ports': []}
                    found_port = None
                    for mapped_port in chassis_lldp_discovery['ports']:
                        if mapped_port['portId'] == port.port_id:
                            found_port = mapped_port

                    if found_port is None:
                        log.info(f"Port {port.port_id} on chassis {port.switch.mac} not in RedSky. Adding.")
                        if simulate is False:
                            self._redsky.add_lldp_discovery(chassis=port.switch.mac,
                                                            ports=[str(port.port_id)],
                                                            location=port_redsky_location,
                                                            description=port.name)
                        audit_results.added_ports.append(port)
                    else:
                        # Check the existing LLDP discovery location against what Meraki says it should be
                        if found_port['erl']['name'] != port_location_tag:
                            log.info(f"Port {device.mac}-{port.port_id} has the wrong location. "
                                     f"Updating to {port_location_tag}")
                            if simulate is False:
                                self._redsky.update_lldp_port_location(entry_id=found_port['id'],
                                                                       chassis_id=chassis_lldp_discovery['id'],
                                                                       port=str(port.port_id),
                                                                       new_location=port_redsky_location,
                                                                       description=port.name)
                            audit_results.updated_locations.append(port)
            elif isinstance(device, MerakiWireless):
                audit_results.access_points.append(device)
                device_location_tag = tags_decoder(device.tags)
                if device_location_tag is None:
                    audit_results.ignored_devices.append(device)
                    continue

                ap_redsky_location = device_building.get_location_by_name(device_location_tag)
                if ap_redsky_location is None:
                    log.info(f"No Location found for {device_location_tag} in Building {device_building.name}. Adding.")
                    if simulate is False:
                        device_building.add_location(location_name=device_location_tag, location_info=device_location_tag)
                        ap_redsky_location = device_building.get_location_by_name(device_location_tag)
                    audit_results.added_locations.append({'building': device_building, 'location': device_location_tag})

                for bss in device.get_bss_list():
                    rs_bssid_found = False
                    for bssid in self._redsky.get_bssid_discovery():
                        if bssid['bssid'].upper() == bss['bssid'].upper():
                            log.info(f"BSSID {bss['bssid']} already in RedSky")
                            rs_bssid_found = True
                    if rs_bssid_found is False:
                        log.info(f"BSSID {bss['bssid']} not in RedSky. Adding")
                        if simulate is False:
                            self._redsky.add_bssid_discovery(bssid=bss['bssid'].upper(),
                                                             location=ap_redsky_location,
                                                             description=f"{bss['ssid_name']} {bss['band']}",
                                                             masking=False)
                        audit_results.added_bssid.append(bss['bssid'])

        return audit_results


class MerakiDevice:
    def __init__(self, dashboard: meraki.DashboardAPI, device_config: dict):
        self.dashboard: meraki.DashboardAPI = dashboard
        """ The API connection to the Meraki Dashboard """
        self.address: str = device_config.get('address', None)
        """ The street address associated with the device """
        self.serial: str = device_config.get('serial', None)
        """ The serial number of the device """
        self.mac: str = device_config.get('mac', None)
        """ The MAC address of the device """
        self.model: str = device_config.get('model', None)
        """ The model name/number of the device """
        self.tags: list[str] = device_config.get('tags', [])
        """ Any tags associated with the device in the Meraki Dashboard """
        self.name: str = device_config.get('name', None)
        """ The name of the device """
        self._redsky: Optional[RedSky] = None
        self._redsky_building = None

    def get_redsky_building(self, redsky: Optional[RedSky] = None):
        """ Use the Meraki Device address to attempt to find the associated Building in RedSky Horizon Mobility.

        Because both addresses are unstructured, this method attempts a "best guess" by normalizing the addresses in
        both systems and trying to make a match. If no match is made, None is returned.

        In order for this method to work, either a :py:class:`wxcadm.Redsky` instance must be passed as an agument or
        the RedSky instance should be attached to the MerakiNetwork with :py:meth:`MerakiNetwork.attach_redsky()`.

        Args:
            redsky (wxcadm.RedSky, optional): The RedSky instance to use in order to call the RedSky APIs

        Returns:
            RedSkyBuilding: The :py:class:`wxcadm.redsky.RedSkyBuilding` instance. None is returned if no match is made.

        """
        if redsky is None:
            if self._redsky is not None:
                redsky = self._redsky
            else:
                raise ValueError("No RedSky instance attached")

        clean_address = address_cleaner(self.address)
        log.debug(f"Finding RedSky Building for: {clean_address}")
        for building in redsky.buildings:
            # Check the full address in Meraki to see if we have a matching Building
            if clean_address == building.address['normalizedAddress']:
                log.debug(f"Full address match: {building.address['normalizedAddress']} - {building.name}")
                self._redsky_building = building
                break
            elif clean_address.split(',')[0] == building.address['streetAddress']:
                log.debug(f"Street address match: {building.address['streetAddress']} - {building.name}")
                self._redsky_building = building
                break
        if self._redsky_building is None:
            log.error(f"No RedSky Building match for {self.address}. Ensure the building is built in RedSky")
        return self._redsky_building


class MerakiSwitch(MerakiDevice):
    def __init__(self, dashboard: meraki.DashboardAPI, device_config: dict):
        super().__init__(dashboard, device_config)
        self._port_list = []

    def get_ports(self):
        """ Get the list of ports on the Switch

        Returns:
            list[MerakiSwitchPort]: A list of :py:class:`MerakiSwitchPort` instances

        """
        port_list = self.dashboard.switch.getDeviceSwitchPorts(serial=self.serial)
        for port in port_list:
            this_port = MerakiSwitchPort(switch=self, port_id=port['portId'], **port)
            self._port_list.append(this_port)
        return self._port_list


class MerakiSwitchPort:
    def __init__(self, switch: MerakiSwitch, port_id: int, **kwargs):
        self.switch = switch
        """ The :py:class:`MerakiSwitch` the port is associated with """
        self.port_id = port_id
        """ The port number """
        self.name: str = kwargs.get('name', "")
        """ The name of the port in the Meraki Dashboard """
        self.tags: list = kwargs.get('tags', [])
        """ A list of tags for the port """
        self.enabled: bool = kwargs.get('enabled', False)
        """ Whether the port is enabled or not """
        self.type: str = kwargs.get('type', 'Unknown')
        """ The type of port """


class MerakiWireless(MerakiDevice):
    def __init__(self, dashboard: meraki.DashboardAPI, device_config: dict):
        super().__init__(dashboard, device_config)
        self._bss_list = []

    def get_bss_list(self):
        """ Get the list of all BSSIDs associated with the Access Point

        Returns:
            list[dict]: A list of the BSS in dict format

        """
        bss_list = self.dashboard.wireless.getDeviceWirelessStatus(serial=self.serial)
        for bss in bss_list['basicServiceSets']:
            if bss['enabled'] is True:
                this_bss = {
                    'bssid': bss['bssid'],
                    'ssid_name': bss['ssidName'],
                    'band': bss['band']
                }
                self._bss_list.append(this_bss)
        return self._bss_list


class MerakiAuditResults:
    def __init__(self):
        self.missing_buildings = []
        """ Buildings that were not found based on the address of the Meraki Device """
        self.added_locations = []
        """ Locations added to existing Buildings in RedSky """
        self.switches = []
        """ List of Meraki switches found in the Dashboard """
        self.access_points = []
        """ List of Meraki Access Points found in the Dashboard """
        self.added_chassis = []
        """ Chassis devices added to RedSky LLDP Discovery """
        self.added_ports = []
        """ Switch ports added to RedSky LLDP Discovery """
        self.ignored_devices = []
        """ Meraki devices that were ignored due to non-present 911 tags """
        self.updated_locations = []
        """ Locations that were uypdated from previous values """
        self.added_bssid = []
        """ BSSIDs added to RedSky BSSID Network Discovery """
        pass



from __future__ import annotations

from typing import Optional, Union
from collections import UserList
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
import json
import requests
import re

import wxcadm.location
from wxcadm import log
from .common import *


@dataclass_json
@dataclass
class RecordingVendor:
    id: str
    """ The ID of the Recording Vendor """
    name: str
    """ The name of the Recording Vendor """
    description: str
    """ The description of the Recording Vendor """
    auto_user_account_creation: bool = field(metadata=config(field_name="migrateUserCreationEnabled"))
    """ Whether user accounts are created automatically on the Recording Vendor """
    login_url: str = field(metadata=config(field_name="loginUrl"))
    """ The URL to access the Recording Provider's login page """
    tos_url: str = field(metadata=config(field_name="termsOfServiceUrl"))
    """ The URL to access the Recording Provider's Terms of Service """


class RecordingVendorsList(UserList):
    def __init__(self, vendors: list):
        super().__init__()
        self.data = []
        for vendor in vendors:
            this_vendor = RecordingVendor.from_dict(vendor)
            self.data.append(this_vendor)

    def get(self, id: Optional[str] = '', name: Optional[str] = '') -> Optional[RecordingVendor]:
        """ Get a :class:`RecordingVendor` by its ID or name

        Args:
            id (str, optional): The ID of the Recording Vendor
            name (str, optional): The name of the Recording Vendor

        Returns:
            RecordingVendor or None

        """
        for vendor in self.data:
            if vendor.id == id or vendor.name == name:
                return vendor
        return None


class OrgRecordingVendorSelection:
    def __init__(self, parent: wxcadm.Org):
        log.info("Getting Org-level Recording Vendor Selection")
        self.parent: wxcadm.Org = parent
        self._url = "v1/telephony/config/callRecording/vendors"
        response = webex_api_call('get', self._url, params={'orgId': self.parent.org_id})
        self.available_vendors: RecordingVendorsList = RecordingVendorsList(response['vendors'])
        """ :class:`RecordingVendorsList` of available :class:`RecordingVendor` """
        self.selected_vendor: RecordingVendor = self.available_vendors.get(id=response['vendorId'])
        """ The selected :class:`RecordingVendor` """
        self.storage_region: Optional[str] = response.get('storageRegion', None)
        """ The storage region. Only applicable when the recording vendor is Webex """
        self.failure_behavior: Optional[str] = response.get('failureBehavior', None)
        """ The behavior when the Call Recording session cannot be established """

    def change_vendor(self, new_vendor: RecordingVendor) -> bool:
        """ Replace the current Recording Vendor with another

        Args:
            new_vendor (RecordingVendor): The new :class:`RecordingVendor`

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'vendorId': new_vendor.id}
        webex_api_call('put', self._url, payload=payload, params={'orgId': self.parent.org_id})
        return True

    def set_failure_behavior(self, failure_behavior: str) -> bool:
        """ Set the failure behavior

        Args:
            failure_behavior (str): The failure behavior. Valid options are ``'PROCEED_WITH_CALL_NO_ANNOUNCEMENT'``,
                ``'PROCEED_CALL_WITH_ANNOUNCEMENT'``, or ``'END_CALL_WITH_ANNOUNCEMENT'``

        Returns:
            bool: True on success, False otherwise

        """


class LocationRecordingVendorSelection:
    def __init__(self, parent: wxcadm.Location):
        log.info("Getting Location-level Recording Vendor Selection")
        self.parent: wxcadm.Location = parent
        self._url = f"v1/telephony/config/locations/{self.parent.id}/callRecording/vendors"
        response = webex_api_call('get', self._url, params={'orgId': self.parent.org_id})
        self.available_vendors: RecordingVendorsList = RecordingVendorsList(response['vendors'])
        """ :class:`RecordingVendorsList` of available :class:`RecordingVendor` """
        self.use_org_vendor: bool = response['orgDefaultEnabled']
        """ Whether to use the Org-level Recording Vendor """
        self.org_vendor = self.available_vendors.get(id=response['orgDefaultVendorId'])
        """ The :class:`RecordingVendor` used at the Org level """
        self.location_vendor = self.available_vendors.get(id=response['defaultVendorId'])
        """ The :class:`RecordingVendor` used at the Location level """
        self.use_org_storage_region: Optional[bool] = response.get('orgStorageRegionEnabled', None)
        """ Whether to use the Org-level Storage Region. Only applies to Webex recording """
        self.org_storage_region: Optional[str] = response.get('orgStorageRegion', None)
        """ The Storage Region used at the Org level """
        self.location_storage_region: Optional[str] = response.get('storageRegion', None)
        """ The Storage Region used at the Location level """
        self.use_org_failure_behavior: bool = response['orgFailureBehaviorEnabled']
        """ Whether to use the Org-level Failure Behavior. """
        self.org_failure_behavior: str = response.get('orgFailureBehavior', None)
        """ The Failure Behavior used at the Org level """
        self.location_failure_behavior: str = response.get('failureBehavior', None)
        """ The Failure Behavior used at the Location level """

    def change_vendor(self, new_vendor: RecordingVendor) -> bool:
        """ Replace the current Recording Vendor with another.

        If the Location is currently using the Recording Vendor from the Org level, the Org-level vendor will no longer
        be used.

        Args:
            new_vendor (RecordingVendor): The new :class:`RecordingVendor`

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'id': new_vendor.id, 'orgDefaultEnabled': False}
        webex_api_call('put', f"v1/telephony/config/locations/{self.parent.id}/callRecording/vendor",
                       payload=payload, params={'orgId': self.parent.org_id})
        self.location_vendor = new_vendor
        return True

    def set_storage_region(self, region: str) -> bool:
        """ Set the Location-level Recording Storage Region (Webex recording only)

        Args:
            region (str): The 2-character region code

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'storageRegion': region, 'orgStorageRegionEnabled': False}
        webex_api_call('put', f"v1/telephony/config/locations/{self.parent.id}/callRecording/vendor",
                       payload=payload, params={'orgId': self.parent.org_id})
        self.use_org_storage_region = False
        self.location_storage_region = region
        return True

    def set_failure_behavior(self, failure_behavior: str) -> bool:
        """ Set the Location-level Recording Failure Behavior

        Args:
            failure_behavior (str): The failure behavior. Valid options are `'PROCEED_WITH_CALL_NO_ANNOUNCEMENT'``,
                ``'PROCEED_CALL_WITH_ANNOUNCEMENT'``, or ``'END_CALL_WITH_ANNOUNCEMENT'``

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'failureBehavior': failure_behavior, 'orgFailureBehaviorEnabled': False}
        webex_api_call('put', f"v1/telephony/config/locations/{self.parent.id}/callRecording/vendor",
                       payload=payload, params={'orgId': self.parent.org_id})
        self.use_org_failure_behavior = False
        self.location_failure_behavior = failure_behavior
        return True

    def clear_vendor_override(self) -> bool:
        """ Revert the Location-level Recording Vendor back to the Org default """
        payload = {'orgDefaultEnabled': True}
        webex_api_call('put', f"v1/telephony/config/locations/{self.parent.id}/callRecording/vendor",
                       payload=payload, params={'orgId': self.parent.org_id})
        self.use_org_vendor = True
        return True

    def clear_region_override(self) -> bool:
        """ Revert the Location-level Storage Region back to the Org default (Webex recording only) """
        payload = {'orgStorageRegionEnabled': True}
        webex_api_call('put', f"v1/telephony/config/locations/{self.parent.id}/callRecording/vendor",
                       payload=payload, params={'orgId': self.parent.org_id})
        self.use_org_storage_region = True
        return True

    def clear_failure_override(self) -> bool:
        """ Revert the Location-level Failure Behavior back to the Org default """
        payload = {'orgFailureBehaviorEnabled': True}
        webex_api_call('put', f"v1/telephony/config/locations/{self.parent.id}/callRecording/vendor",
                       payload=payload, params={'orgId': self.parent.org_id})
        self.use_org_failure_behavior = True
        return True


class ComplianceAnnouncementSettings:
    def __init__(self, parent: wxcadm.Org | wxcadm.Location,
                 inboundPSTNCallsEnabled: bool,
                 outboundPSTNCallsEnabled: bool,
                 outboundPSTNCallsDelayEnabled: bool,
                 delayInSeconds: int,
                 useOrgSettingsEnabled: Optional[bool] = None):
        self.parent: wxcadm.Org | wxcadm.Location = parent
        """ The :class:`Org` or :class:`Location` that the settings apply to """
        self.inbound_pstn_calls_enabled: bool = inboundPSTNCallsEnabled
        """ Play compliance announcement for inbound PSTN calls """
        self.outbound_pstn_calls_enabled: bool = outboundPSTNCallsEnabled
        """ Play compliance announcement for outbound PSTN calls """
        self.outbound_pstn_calls_delay_enabled: bool = outboundPSTNCallsDelayEnabled
        """ Delay the compliance announcement for a number of seconds for outbound PSTN calls """
        self.delay: int = delayInSeconds
        """ The number of seconds to delay the message for outbound calls, if enabled """
        self.use_org_settings: Optional[bool] = useOrgSettingsEnabled
        """ For Location-level Compliance Announcements, whether to override the Org settings (False) """

    def to_webex(self) -> dict:
        """ Represent the instance as a dict with the Webex field names as keys

        Returns:
            dict: The dict of Webex-friendly field names::

                {
                    'inboundPSTNCallsEnabled': self.inbound_pstn_calls_enabled,
                    'outboundPSTNCallsEnabled': self.outbound_pstn_calls_enabled,
                    'outboundPSTNCallsDelayEnabled': self.outbound_pstn_calls_delay_enabled,
                    'delayInSeconds': self.delay,
                    'useOrgSettingsEnabled': self.use_org_settings
                }

        """
        webex_dict = {
            'inboundPSTNCallsEnabled': self.inbound_pstn_calls_enabled,
            'outboundPSTNCallsEnabled': self.outbound_pstn_calls_enabled,
            'outboundPSTNCallsDelayEnabled': self.outbound_pstn_calls_delay_enabled,
            'delayInSeconds': self.delay,
            'useOrgSettingsEnabled': self.use_org_settings
        }
        return webex_dict

    def to_json(self) -> str:
        """ Represent the instance as a JSON string

        Returns:
            str: The JSON representation of the configuration

        """
        return json.dumps(self.to_webex())

    def push(self) -> bool:
        """ Push any changes (the existing instance config) back to Webex

        Returns:
            bool: True on success, False otherwise

        """
        if isinstance(self.parent, wxcadm.Org):
            pass
        pass


class Recording:
    """ The class for Converged Recordings """
    def __init__(self, parent: wxcadm.Org, id: str, details: Optional[dict] = None, timezone: Optional[str] = None):
        self._timezone = timezone
        self.parent = parent
        self.id = id
        """ The Recording ID """
        self.details: dict
        """ The full details of the Recording from Webex """
        if details is None:
            self.details = self._get_details()
        else:
            self.details = details

    def _get_details(self):
        params = {'timezone': self._timezone} if self._timezone else None
        response = webex_api_call('get', f'v1/convergedRecordings/{self.id}', params=params)
        return response

    def refresh(self) -> None:
        self.details = self._get_details()



    @property
    def status(self) -> str:
        """ The status of the Recording """
        status = self.details.get('status', self._get_details()['status'])
        return status

    @property
    def url(self) -> str:
        """ The URL to download the recording """
        download = self.details.get('temporaryDirectDownloadLinks', self._get_details()['temporaryDirectDownloadLinks'])
        log.debug(download)
        return download['audioDownloadLink']

    @property
    def transcript_url(self) -> str:
        """ The URL to download the recording transcript """
        download = self.details.get('temporaryDirectDownloadLinks', self._get_details()['temporaryDirectDownloadLinks'])
        return download['transcriptDownloadLink']

    @property
    def suggested_notes_url(self) -> str:
        """ The URL to download the recording suggested notes """
        download = self.details.get('temporaryDirectDownloadLinks', self._get_details()['temporaryDirectDownloadLinks'])
        return download['suggestedNotesDownloadLink']

    @property
    def action_items_url(self) -> str:
        """ The URL to download the recording action items """
        download = self.details.get('temporaryDirectDownloadLinks', self._get_details()['temporaryDirectDownloadLinks'])
        return download['actionItemsDownloadLink']

    @property
    def short_notes_url(self) -> str:
        """ The URL to download the recording short notes """
        download = self.details.get('temporaryDirectDownloadLinks', self._get_details()['temporaryDirectDownloadLinks'])
        return download['shortNotesDownloadLink']

    @property
    def download_expires(self) -> str:
        """ The expiration date and time of the :attr:`url` """
        download = self.details.get('temporaryDirectDownloadLinks', self._get_details()['temporaryDirectDownloadLinks'])
        return download['expiration']

    @property
    def file_format(self) -> str:
        """ The format of the recording file """
        file_format = self.details.get('format', self._get_details()['format'])
        return file_format

    @property
    def duration(self) -> int:
        """ The duration of the recording, in seconds """
        duration = self.details.get('durationSeconds', self._get_details()['durationSeconds'])
        return duration

    @property
    def file_size(self) -> int:
        """ The size of the recording file, in bytes """
        file_size = self.details.get('sizeBytes', self._get_details()['sizeBytes'])
        return file_size

    @property
    def topic(self) -> str:
        """ The Topic or Description of the Recording """
        topic = self.details.get('topic', "")
        return topic

    @property
    def service_type(self) -> str:
        """ The Service Type that created the Recording """
        service_type = self.details.get('serviceType', self._get_details()['serviceType'])
        return service_type

    @property
    def storage_region(self) -> str:
        """ The region where the Recording is stored """
        region = self.details.get('storageRegion', self._get_details()['storageRegion'])
        return region

    @property
    def created(self) -> str:
        """ The date and time the recording file was created """
        created = self.details.get('createTime', self._get_details()['createTime'])
        return created

    @property
    def recorded(self) -> str:
        """ The date and time the call was recorded """
        recorded = self.details.get('timeRecorded', self._get_details()['timeRecorded'])
        return recorded

    @property
    def owner_id(self) -> str:
        """ The UUID of the :class:`Person`, :class:`Workspace` or :class:`VirtualLine` that was recorded """
        owner_id = self.details.get('ownerId', self._get_details()['ownerId'])
        return owner_id

    @property
    def owner_type(self) -> str:
        """ The type (user, workspace, virtual line) of the :attr:`owner_id` """
        owner_type = self.details.get('ownerType', self._get_details()['ownerType'])
        return owner_type

    @property
    def owner_email(self) -> str:
        """ The email address of the owner, if owned by something that supports email address """
        owner_email = self.details.get('ownerEmail', self._get_details()['ownerEmail'])
        return owner_email

    @property
    def location_id(self) -> Optional[str]:
        """ For Calling service recordings, the Location ID of the :attr:`owner_id` """
        if self.service_type.lower() != 'calling':
            return None
        location_id = self.details['serviceData']['locationId']
        return location_id

    @property
    def call_session_id(self) -> Optional[str]:
        """ For Calling service recordings, the Call Session ID of the Recording """
        if self.service_type.lower() != 'calling':
            return None
        session = self.details['serviceData']['callSessionId']
        return session

    def download(self, filename: str) -> bool:
        """ Download the recording file to the local machine

        Args:
            filename (str): The filename to save the Recording to

        Returns:
            bool: True on success

        """
        log.debug(f"Downloading recording to file: {filename}")
        response = requests.get(self.url)
        with open(filename, 'wb') as f:
            f.write(response.content)
        return True

    def get_transcript(self, text_only: bool = False) -> str:
        """ Get the transcript of the call

        Args:
            text_only (bool, optional): Return only the transcript without timestamps if True

        Returns:
            str: The transcript of the call

        """
        log.debug(f"Getting transcript for recording: {self.id}")
        response = requests.get(self.transcript_url)
        if text_only is True:
            log.debug("Parsing transcript to text only")
            raw_transcript = []
            text = re.compile("[A-Za-z]")
            log.debug(f"Parsing transcript: {response.text}")
            lines = response.text.splitlines()
            for line in lines:
                if re.match(text, line) and line != 'WEBVTT':
                    raw_transcript.append(line)
            return " ".join(raw_transcript)
        else:
            return response.text

    def get_suggested_notes(self):
        response = requests.get(self.suggested_notes_url)
        return response.text

    def get_action_items(self):
        response = requests.get(self.action_items_url)
        return response.text

    def get_short_notes(self):
        response = requests.get(self.short_notes_url)
        return response.text



class RecordingList(UserList):
    _endpoint = 'v1/admin/convergedRecordings'
    _endpoint_items_key = None
    _item_endpoint = 'v1/convergedRecordings/{item_id}'
    _item_class = Recording

    def __init__(self, parent: wxcadm.Org,
                 from_date_time: Optional[str] = None,
                 to_date_time: Optional[str] = None,
                 status: Optional[str] = None,
                 service: Optional[str] = None,
                 owner: Optional[Union[wxcadm.Person, wxcadm.Workspace, wxcadm.VirtualLine]] = None,
                 region: Optional[str] = None,
                 location: Optional[wxcadm.Location] = None
                 ):
        super().__init__()
        log.debug("Initializing RecordingList")
        self.parent: wxcadm.Org = parent
        self.params = {'max': 100}
        if from_date_time is not None:
            self.params['from'] = str(from_date_time)
        if to_date_time is not None:
            self.params['to'] = str(to_date_time)
        if status is not None:
            self.params['status'] = str(status)
        if service is not None:
            self.params['serviceType'] = str(service)
        if owner is not None:
            self.params['ownerId'] = str(owner.id)
        if region is not None:
            self.params['storageRegion'] = str(region)
        if location is not None:
            self.params['locationId'] = location.id
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        log.debug(f"_get_data() started with params {self.params}")
        response = webex_api_call('get', self._endpoint, params=self.params)

        items = []
        for entry in response:
            items.append(self._item_class(parent=self.parent, id=entry['id'], details=entry))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self, id: str):
        """ Get the instance associated with a given Recording ID

        Args:
            id (str): The Recording ID to find

        Returns:
            Recording: The :class:`Recording` instance correlating with the given ID

        """
        for item in self.data:
            if item.id == id:
                return item
        return None
